/* $LAAS: format.c 2010/02/25 16:28:30 mallet $ */

/*
 * Copyright (c) 2008,2010 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 *
 *					Anthony Mallet on Fri Jun 27 2008
 */

static const char _rcsid[] = "$LAAS: format.c 2010/02/25 16:28:30 mallet $";

#include <stdio.h>
#include <string.h>
#include <err.h>

#include "viam_private.h"

#ifdef VIAM_OPENCV_SUPPORT
# ifndef CV_INLINE
#  define CV_INLINE	static inline
# endif
# define CV_NO_BACKWARD_COMPATIBILITY
# include <cxtypes.h>
# include <cv.h>
#endif


/* --- local data ---------------------------------------------------- */

#define clamp(v)	(((v)<0.)?0.:(((v)>255.)?255.:(v)))

#define YUV_R(y, u, v)	clamp((y) + 1.14 * (v))
#define YUV_G(y, u, v)	clamp((y) - 0.395 * (u) - 0.581 * (v))
#define YUV_B(y, u, v)	clamp((y) + 2.032 * (u))

static int	viam_fmt_demux(viam_handle_t vh, viam_filterformat_t format,
			const IplImage *source, IplImage *result);
static int	viam_yuv422_to_bgr(viam_handle_t vh, const IplImage *source,
			IplImage *result);
static int	viam_yuv411_to_bgr(viam_handle_t vh, const IplImage *source,
			IplImage *result);
static int	viam_yuv444_to_bgr(viam_handle_t vh, const IplImage *source,
			IplImage *result);


/* --- viam_format_geosetup ------------------------------------------ */

/** \brief Compute geometry of format filters.
 *
 * \return 0 on success.
 */

int
viam_format_geosetup(viam_handle_t vh, viam_image_ref image,
		     viam_filterargs_obj args, viam_geometry_t *geom,
		     void **data)
{
  switch(args->format) {
    case VIAM_FILTER_DEMUX_EVENPX:
    case VIAM_FILTER_DEMUX_ODDPX:
      /* geometry is unchanged */
      break;

#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_RGB_TO_BGR:
      /* check validity */
      if (geom->channels != 3) {
	warnx("color filters require three channels image");
	return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      }

      /* geometry is unchanged */
      break;

    case VIAM_FILTER_RGB_TO_GRAY:
    case VIAM_FILTER_BGR_TO_GRAY:
      /* check validity */
      if (geom->channels != 3) {
	warnx("color to gray filters require 3 channels image");
	return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      }

      /* update channels */
      geom->channels = 1;
      break;
#endif /* VIAM_OPENCV_SUPPORT */

#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_BAYER_BGGR:
    case VIAM_FILTER_BAYER_GBRG:
    case VIAM_FILTER_BAYER_RGGB:
    case VIAM_FILTER_BAYER_GRBG:
#endif /* VIAM_OPENCV_SUPPORT */
    case VIAM_FILTER_YUV422_TO_BGR:
    case VIAM_FILTER_YUV411_TO_BGR:
    case VIAM_FILTER_YUV444_TO_BGR:
      if (geom->channels != 1) {
	warnx("gray to color filters require single channels image");
	return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
      }

      /* update channels */
      geom->channels = 3;
      break;
  }

  return 0;
}


/* --- viam_cvtformat ------------------------------------------------ */

/** \brief Transform image from some format to some other.
 *
 * \param[in] vh	Library handle.
 * \param[in] image	Image object of the filter
 * \param[in] source	Source image
 * \param[out] result	Resulting image
 * \param[in] args	Filter arguments
 * \param[in] data	Filter private data
 *
 * \return 0 on success.
 */

int
viam_cvtformat(viam_handle_t vh, viam_image_ref image, const IplImage *source,
	       IplImage *result, viam_filterargs_obj args, void *data)
{
  int s;

  switch(args->format) {
    case VIAM_FILTER_DEMUX_EVENPX:
    case VIAM_FILTER_DEMUX_ODDPX:
      s = viam_fmt_demux(vh, args->format, source, result);
      break;

#ifdef VIAM_OPENCV_SUPPORT
    case VIAM_FILTER_RGB_TO_BGR:
    case VIAM_FILTER_RGB_TO_GRAY:
    case VIAM_FILTER_BGR_TO_GRAY:
    case VIAM_FILTER_BAYER_BGGR:
    case VIAM_FILTER_BAYER_GBRG:
    case VIAM_FILTER_BAYER_RGGB:
    case VIAM_FILTER_BAYER_GRBG: {

      static const int code[] = {
	[VIAM_FILTER_RGB_TO_BGR] =	CV_RGB2BGR,
	[VIAM_FILTER_RGB_TO_GRAY] =	CV_RGB2GRAY,
	[VIAM_FILTER_BGR_TO_GRAY] =	CV_BGR2GRAY,
	[VIAM_FILTER_BAYER_BGGR] =	CV_BayerBG2BGR,
	[VIAM_FILTER_BAYER_GBRG] =	CV_BayerGB2BGR,
	[VIAM_FILTER_BAYER_RGGB] =	CV_BayerRG2BGR,
	[VIAM_FILTER_BAYER_GRBG] =	CV_BayerGR2BGR
      };

      cvCvtColor(source, result, code[args->format]);
      if (cvGetErrStatus())
	s = viam_seterrno(vh, Eviamlib_OPENCV);
      else
	s = 0;
      break;
    }
#endif /* VIAM_OPENCV_SUPPORT */

    case VIAM_FILTER_YUV422_TO_BGR:
      s = viam_yuv422_to_bgr(vh, source, result);
      break;

    case VIAM_FILTER_YUV411_TO_BGR:
      s = viam_yuv411_to_bgr(vh, source, result);
      break;

    case VIAM_FILTER_YUV444_TO_BGR:
      s = viam_yuv444_to_bgr(vh, source, result);
      break;
  }

  return s;
}


/* --- viam_fmt_demux ------------------------------------------------ */

/** \brief Demultiplex image.
 *
 * \param[in] vh	Library handle
 * \param[in] format	Type of demultiplexing
 * \param[in] source	Source image
 * \param[out] result	Resulting image
 *
 * \return 0 on success.
 */

static int
viam_fmt_demux(viam_handle_t vh, viam_filterformat_t format,
	       const IplImage *source, IplImage *result)
{
  unsigned char *dl, *sl, *dp, *sp;
  size_t size, offset;
  int i,j;

  switch(format) {
    case VIAM_FILTER_DEMUX_EVENPX:
      offset = 0;
      break;

    case VIAM_FILTER_DEMUX_ODDPX:
      offset = (source->depth>>3)*source->nChannels;
      break;

    default: return viam_seterrno(vh, Eviamlib_INVALID_VALUE);
  }

  size = source->widthStep/2;
  if (size > source->width) size = source->width;
  if (size > result->width) size = result->width;

  sl = (unsigned char *)source->imageData;
  dl = (unsigned char *)result->imageData;
  for(i=0; i<result->height; i++) {

    dp = dl; sp = sl;
    for(j=0; j<size; j++) {
      memcpy(dp, sp + offset, (source->depth>>3)*source->nChannels);
      dp += (result->depth>>3)*result->nChannels;
      sp += 2*(source->depth>>3)*source->nChannels;
    }

    sl += source->widthStep;
    dl += result->widthStep;
  }

  return 0;
}


/* --- viam_yuv422_to_bgr -------------------------------------------- */

/** \brief Transform image from YUV422 to BGR representation.
 *
 * Pixels in YUV422 format are arranged like this:
 *   u, y1, v, y2...
 * BGR representation is done like this: 
 *   bgr1 = RGB(y1, u, v)
 *   bgr2 = RGB(y2, u, v) ...
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viam_yuv422_to_bgr(viam_handle_t vh, const IplImage *source, IplImage *result)
{
  unsigned char *s, *d;
  unsigned int i,j;
  int y1, y2, u, v;

  s = (unsigned char *)source->imageData;
  d = (unsigned char *)result->imageData;
  for(i = 0; i<source->height; i++) {
    for(j = 0; j<source->widthStep/4; j++) {
      u = *(s++); y1 = *(s++); v = *(s++); y2 = *(s++);
      u -= 128; v -= 128;

      *(d++) = YUV_B(y1, u, v);
      *(d++) = YUV_G(y1, u, v);
      *(d++) = YUV_R(y1, u, v);

      *(d++) = YUV_B(y2, u, v);
      *(d++) = YUV_G(y2, u, v);
      *(d++) = YUV_R(y2, u, v);
    }
  }

  return 0;
}


/* --- viam_yuv411_to_bgr -------------------------------------------- */

/** \brief Transform image from YUV411 to BGR representation.
 *
 * Pixels in YUV411 format are arranged like this:
 *   u, y1, y2, v, y3, y4...
 * BGR representation is done like this: 
 *   bgr1 = RGB(y1, u, v)
 *   bgr2 = RGB(y2, u, v) ...
 *   bgr3 = RGB(y3, u, v) ...
 *   bgr4 = RGB(y4, u, v) ...
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viam_yuv411_to_bgr(viam_handle_t vh, const IplImage *source, IplImage *result)
{
  unsigned char *s, *d;
  unsigned int i,j;
  int y1, y2, y3, y4, u, v;

  s = (unsigned char *)source->imageData;
  d = (unsigned char *)result->imageData;
  for(i = 0; i<source->height; i++) {
    for(j = 0; j<source->widthStep/6; j++) {
      u = *(s++); y1 = *(s++); y2 = *(s++);
      v = *(s++); y3 = *(s++); y4 = *(s++);
      u -= 128; v -= 128;

      *(d++) = YUV_B(y1, u, v);
      *(d++) = YUV_G(y1, u, v);
      *(d++) = YUV_R(y1, u, v);

      *(d++) = YUV_B(y2, u, v);
      *(d++) = YUV_G(y2, u, v);
      *(d++) = YUV_R(y2, u, v);

      *(d++) = YUV_B(y3, u, v);
      *(d++) = YUV_G(y3, u, v);
      *(d++) = YUV_R(y3, u, v);

      *(d++) = YUV_B(y4, u, v);
      *(d++) = YUV_G(y4, u, v);
      *(d++) = YUV_R(y4, u, v);
    }
  }

  return 0;
}


/* --- viam_yuv444_to_bgr -------------------------------------------- */

/** \brief Transform image from YUV444 to BGR representation.
 *
 * Pixels in YUV444 format are arranged like this:
 *   uyv1, uyv2, ...
 * BGR representation is done like this: 
 *   bgr1 = RGB(y1, u1, v1)
 *   bgr2 = RGB(y2, u2, v2) ...
 *
 * \param[in] vh	Library handle.
 *
 * \return 0 on success.
 */

static int
viam_yuv444_to_bgr(viam_handle_t vh, const IplImage *source, IplImage *result)
{
  unsigned char *s, *d;
  unsigned int i,j;
  int y, u, v;

  s = (unsigned char *)source->imageData;
  d = (unsigned char *)result->imageData;
  for(i = 0; i<source->height; i++) {
    for(j = 0; j<source->widthStep/3; j++) {
      u = *(s++); y = *(s++); v = *(s++);
      u -= 128; v -= 128;

      *(d++) = YUV_B(y, u, v);
      *(d++) = YUV_G(y, u, v);
      *(d++) = YUV_R(y, u, v);
    }
  }

  return 0;
}
